"""
HAL JSON link modeling.

HAL JSON defines a simple-yet-powerful extension to plain old JSON that allows
resources to link to each other, enabling HATEOAS and the full power of REST.

See: https://tools.ietf.org/html/draft-kelly-json-hal-07

"""
from werkzeug.routing import BuildError

from microcosm_flask.namespaces import Namespace

# NB: it would be nice to use marshmallow schemas in lieu of `to_dict()` functions here
#
# The main obstacles are:
#
#  - Links values can be dictionaries (with an "href") or lists of the same.
#  - Links are encoded in several places; it's a bit more convenient to call `to_dict()`
#    than to instantiate a schema instance in each case.


class Links(object):
    """
    A collection of links organized by relation name.

    """
    def __init__(self, links=None, **kwargs):
        self.links = links or {}
        for key, value in kwargs.items():
            self[key] = value

    def __getitem__(self, name):
        return self.links[name]

    def __setitem__(self, name, value):
        self.links[name] = value

    def to_dict(self):
        def encode(value):
            if isinstance(value, list):
                return [item.to_dict() for item in value]
            else:
                return value.to_dict()

        return {
            name: encode(value)
            for name, value in self.links.items()
        }


class Link(object):
    """
    A single link for a relation name.

    """
    def __init__(self, href, type=None, templated=False):
        self.href = href
        self.type = type
        self.templated = templated

    def to_dict(self):
        dct = {
            "href": self.href,
        }
        if self.type is not None:
            dct["type"] = self.type
        if self.templated:
            dct["templated"] = True
        return dct

    @classmethod
    def for_(cls, operation, ns, qs=None, type=None, allow_templates=False, **kwargs):
        """
        Create a link to an operation on a resource object.

        Supports link templating if enabled by making a best guess as to the URI
        template construction.

        See also [RFC 6570]( https://tools.ietf.org/html/rfc6570).

        :param operation: the operation
        :param ns: the namespace
        :param qs: an optional query string (e.g. for paging)
        :param type: an optional link type
        :param allow_templates: whether generated links are allowed to contain templates
        :param kwargs: optional endpoint expansion arguments (e.g. for URI parameters)
        :raises BuildError: if link templating is needed and disallowed
        """
        # ensure that we actually have a Namespace; many legacy code paths use strings or tuples
        ns = Namespace.make(ns)

        try:
            href, templated = ns.href_for(operation, qs=qs, **kwargs), False
        except BuildError as error:
            if not allow_templates:
                raise
            uri_templates = {
                argument: "{{{}}}".format(argument)
                for argument in error.suggested.arguments
            }
            kwargs.update(uri_templates)
            href, templated = ns.href_for(operation, qs=qs, **kwargs), True

        return cls(
            href=href,
            type=type,
            templated=templated,
        )



"""
Custom fields.

"""
from datetime import datetime

from dateutil import parser
from marshmallow.fields import Field, ValidationError


class EnumField(Field):
    """
    Enum valued field.

    Cribbed from [marshmallow_enum](https://github.com/justanr/marshmallow_enum), which
    is not available on PyPi.

    """
    default_error_messages = {
        'by_name': 'Invalid enum member {name}',
        'by_value': 'Invalid enum value {value}'
    }

    def __init__(self, enum, by_value=False, *args, **kwargs):
        self.enum = enum
        self.by_value = by_value
        super(EnumField, self).__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj):
        if value is None:
            return value
        elif self.by_value:
            return value.value
        else:
            return value.name

    def _deserialize(self, value, attr, data):
        if value is None:
            return value
        elif self.by_value:
            return self._deserialize_by_value(value, attr, data)
        else:
            return self._deserialize_by_name(value, attr, data)

    def _deserialize_by_value(self, value, attr, data):
        try:
            return self.enum(value)
        except ValueError:
            try:
                return self.enum(int(value))
            except:
                pass
            self.fail('by_value', value=value)

    def _deserialize_by_name(self, value, attr, data):
        try:
            return getattr(self.enum, value)
        except AttributeError:
            self.fail('by_name', name=value)


class TimestampField(Field):
    """
    Timestamp valued field, as either a unix timestamp (default) or isoformat string.

    """
    EPOCH = datetime(1970, 1, 1)

    def __init__(self, use_isoformat=False, *args, **kwargs):
        self.use_isoformat = use_isoformat
        super(TimestampField, self).__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj):
        """
        Serialize value as a timestamp, either as a Unix timestamp (in float second) or a UTC isoformat string.

        """
        if value is None:
            return None

        if self.use_isoformat:
            return datetime.utcfromtimestamp(value).isoformat()
        else:
            return value

    def _deserialize(self, value, attr, obj):
        """
        Deserialize value as a Unix timestamp (in float seconds).

        Handle both numeric and UTC isoformat strings.

        """
        if value is None:
            return None

        try:
            return float(value)
        except ValueError:
            parsed = parser.parse(value)
            if parsed.tzinfo:
                if parsed.utcoffset().total_seconds():
                    raise ValidationError("Timestamps must be defined in UTC")
                parsed = parsed.replace(tzinfo=None)
            return (parsed - TimestampField.EPOCH).total_seconds()



"""
Sync entry point.

Sync attempts to read data from either a server (via JSON over HTTP) or a file (as YAML)
and write the same data to another (possibly the same) server or file.

"""
from argparse import ArgumentParser
from logging import basicConfig, DEBUG, ERROR, getLogger, INFO, WARN
import re

from microcosm_flask.sync.pull import pull
from microcosm_flask.sync.push import push
from microcosm_flask.sync.toposort import toposorted


def parse_args():
    parser = ArgumentParser()
    verbosity_group = parser.add_mutually_exclusive_group()
    verbosity_group.add_argument(
        "--verbose",
        "-v",
        action="count",
        dest="verbosity",
        default=1,
        help="Set verbosity level",
    )
    verbosity_group.add_argument(
        "--quiet",
        "-q",
        action="store_const",
        const=0,
        dest="verbosity",
        help="Suppress verbosity",
    )
    follow_group = parser.add_mutually_exclusive_group()
    follow_group.add_argument(
        "--all",
        action="store_true",
        help="Traverse all links",
    )
    follow_group.add_argument(
        "--follow",
        action="append",
        help="Traverse links following one or more patterns",
    )
    parser.add_argument(
        "--exclude-first",
        "-x",
        action="store_true",
        help="Exclude the first resource from pull output",
    )
    parser.add_argument(
        "--batch-size",
        "-b",
        type=int,
        default=1,
        help="Batch size (requires backend PATCH support if >1)",
    )
    parser.add_argument(
        "--enable-sessions",
        action="store_true",
    )
    parser.add_argument(
        "input",
        help="Input location for resources",
    )
    parser.add_argument(
        "output",
        help="Output location for resources",
    )
    return parser.parse_args()


def set_relation_patterns(args):
    if args.all:
        patterns = [".*"]
    elif args.follow:
        patterns = args.follow
    else:
        patterns = ["search", "next"]

    args.relation_patterns = [
        re.compile(pattern) for pattern in patterns
    ]


def set_verbosity(args):
    verbosity = min(args.verbosity, 3)
    level = {0: ERROR, 1: WARN, 2: INFO, 3: DEBUG}.get(verbosity, WARN)
    basicConfig(level=level, format="%(name)s - %(message)s")
    getLogger("sync").setLevel(level)
    if verbosity < 3:
        getLogger("requests.packages.urllib3.connectionpool").setLevel(WARN)


def main():
    args = parse_args()
    set_relation_patterns(args)
    set_verbosity(args)
    data = pull(args)
    push(args, toposorted(data))



"""
Routing registration support.

Intercepts Flask's normal route registration to inject conventions.

"""
from flask_cors import cross_origin

from microcosm.api import defaults


def make_path(graph, path):
    return graph.config.route.path_prefix + path


@defaults(
    converters=[
        "uuid",
    ],
    enable_audit=True,
    enable_basic_auth=False,
    enable_cors=True,
    path_prefix="/api",
)
def configure_route_decorator(graph):
    """
    Configure a flask route decorator that operates on `Operation` and `Namespace` objects.

    By default, enables CORS support, assuming that service APIs are not exposed
    directly to browsers except when using API browsing tools.

    Usage:

        @graph.route(ns.collection_path, Operation.Search, ns)
        def search_foo():
            pass

    """
    # routes depends on converters
    graph.use(*graph.config.route.converters)

    def route(path, operation, ns):
        """
        :param path: a URI path, possibly derived from a property of the `ns`
        :param operation: an `Operation` enum value
        :param ns: a `Namespace` instance
        """
        def decorator(func):
            if graph.config.route.enable_cors:
                func = cross_origin(supports_credentials=True)(func)

            if graph.config.route.enable_basic_auth:
                func = graph.basic_auth.required(func)

            # keep audit decoration last (before registering the route) so that
            # errors raised by other decorators are captured in the audit trail
            if graph.config.route.enable_audit:
                func = graph.audit(func)

            graph.app.route(
                make_path(graph, path),
                endpoint=ns.endpoint_for(operation),
                methods=[operation.value.method],
            )(func)
            return func
        return decorator
    return route



"""
Audit log support for Flask routes.

"""
from collections import namedtuple
from functools import wraps
from logging import getLogger
from json import loads
from traceback import format_exc

from flask import current_app, g, request
from microcosm.api import defaults
from microcosm_flask.errors import (
    extract_context,
    extract_error_message,
    extract_status_code,
)


X_REQUEST = "X-Request"

AuditOptions = namedtuple("AuditOptions", [
    "include_request_body",
    "include_response_body",
    "include_header",
])


def include_header_matching_prefix(prefix):
    """
    Match all headers starting with prefix.

    """
    def include_header(header):
        return header.startswith(prefix)
    return include_header


def audit(func):
    """
    Record a Flask route function in the audit log.

    Generates a JSON record in the Flask log for every request.

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        options = AuditOptions(
            include_request_body=True,
            include_response_body=True,
            include_header=include_header_matching_prefix(X_REQUEST),
        )
        return _audit_request(options, func, *args, **kwargs)

    return wrapper


def _audit_request(options, func, *args, **kwargs):
    """
    Run a request function under audit.

    """
    logger = getLogger("audit")

    response = None

    # always include these fields
    audit_dict = dict(
        operation=request.endpoint,
        func=func.__name__,
        method=request.method,
    )

    # include request body on debug (if any)
    if all((
        current_app.debug,
        options.include_request_body,
        request.get_json(force=True, silent=True),
        not g.get("hide_body"),
    )):
        audit_dict["request_body"] = request.get_json(force=True)

    # include headers (conditionally)
    audit_dict.update({
        header: value
        for header, value in request.headers.items()
        if options.include_header(header)
    })

    # process the request
    try:
        response = func(*args, **kwargs)
    except Exception as error:
        status_code = extract_status_code(error)
        success = 0 < status_code < 400
        audit_dict.update(
            success=success,
            message=extract_error_message(error)[:2048],
            context=extract_context(error),
            stack_trace=None if success else format_exc(limit=10),
            status_code=status_code,
        )
        raise
    else:
        body, status_code = parse_response(response)

        audit_dict.update(
            success=True,
            status_code=status_code,
        )

        # include response body on debug (if any)
        if all((
                current_app.debug,
                options.include_response_body,
                body,
                not g.get("hide_body"),
        )):
            try:
                audit_dict["response_body"] = loads(body)
            except (TypeError, ValueError):
                # not json
                audit_dict["response_body"] = body

        return response
    finally:
        # always log at INFO; we can't know whether a raised exception
        # is an error or expected behavior
        logger.info(audit_dict)


def parse_response(response):
    """
    Parse a Flask response into a body and status code.

    The returned value from a Flask view could be:
        * a tuple of (response, status) or (response, status, headers)
        * a Response object
        * a string
    """
    if isinstance(response, tuple) and len(response) > 1:
        return response[0], response[1]
    try:
        return response.data, response.status_code
    except AttributeError:
        return response, 200


@defaults(
    include_request_body=True,
    include_response_body=True,
    include_header_prefix=X_REQUEST,
)
def configure_audit_decorator(graph):
    """
    Configure the audit decorator.

    Example Usage:

        @graph.audit
        def login(username, password):
            ...
    """
    include_request_body = graph.config.audit.include_request_body
    include_response_body = graph.config.audit.include_response_body
    include_header_prefix = graph.config.audit.include_header_prefix

    def _audit(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            options = AuditOptions(
                include_request_body=include_request_body,
                include_response_body=include_response_body,
                include_header=include_header_matching_prefix(include_header_prefix),
            )
            return _audit_request(options, func, *args, **kwargs)

        return wrapper
    return _audit



"""
Convention base class.

"""
from microcosm_flask.operations import Operation


class EndpointDefinition(tuple):
    """
    A definition for an endpoint.

    """
    def __new__(cls, func=None, request_schema=None, response_schema=None):
        """
        :param func: a function to process request data and return response data
        :param request_schema: a marshmallow schema to decode/validate request data
        :param response_schema: a marshmallow schema to encode response data
        """
        return tuple.__new__(EndpointDefinition, (func, request_schema, response_schema))

    @property
    def func(self):
        return self[0]

    @property
    def request_schema(self):
        return self[1]

    @property
    def response_schema(self):
        return self[2]


class Convention(object):
    """
    A convention is a recipe for applying Flask-compatible functions to a namespace.

    """
    def __init__(self, graph):
        self.graph = graph

    def configure(self, ns, mappings=None, **kwargs):
        """
        Apply mappings to a namespace.

        """
        if mappings is None:
            mappings = dict()
        mappings.update(kwargs)

        for operation, definition in mappings.items():

            try:
                configure_func = self._find_func(operation)
            except AttributeError:
                pass
            else:
                configure_func(ns, self._make_definition(definition))

    def _find_func(self, operation):
        """
        Find the function to use to configure the given operation.

        The input might be an `Operation` enum or a string.

        """
        if isinstance(operation, Operation):
            operation_name = operation.name.lower()
        else:
            operation_name = operation.lower()

        return getattr(self, "configure_{}".format(operation_name))

    def _make_definition(self, definition):
        """
        Generate a definition.

        The input might already be a `EndpointDefinition` or it might be a tuple.

        """
        if not definition:
            return EndpointDefinition()
        if isinstance(definition, EndpointDefinition):
            return definition
        elif len(definition) == 1:
            return EndpointDefinition(
                func=definition[0],
            )
        elif len(definition) == 2:
            return EndpointDefinition(
                func=definition[0],
                response_schema=definition[1],
            )
        elif len(definition) == 3:
            return EndpointDefinition(
                func=definition[0],
                request_schema=definition[1],
                response_schema=definition[2],
            )



"""
Adapter between conventional crud functions and the `microcosm_postgres.store.Store` interface.

"""
from microcosm_flask.naming import name_for


class CRUDStoreAdapter(object):
    """
    Adapt the CRUD conventions callbacks to the `Store` interface.

    Does NOT impose transactions; use the `microcosm_postgres.context.transactional` decorator.

    """
    def __init__(self, graph, store):
        self.graph = graph
        self.store = store

    @property
    def identifier_key(self):
        return "{}_id".format(name_for(self.store.model_class))

    def create(self, **kwargs):
        model = self.store.model_class(**kwargs)
        return self.store.create(model)

    def delete(self, **kwargs):
        identifier = kwargs.pop(self.identifier_key)
        return self.store.delete(identifier)

    def replace(self, **kwargs):
        identifier = kwargs.pop(self.identifier_key)
        model = self.store.model_class(id=identifier, **kwargs)
        return self.store.replace(identifier, model)

    def retrieve(self, **kwargs):
        identifier = kwargs.pop(self.identifier_key)
        return self.store.retrieve(identifier)

    def search(self, offset, limit, **kwargs):
        items = self.store.search(offset=offset, limit=limit, **kwargs)
        count = self.store.count(**kwargs)
        return items, count

    def update(self, **kwargs):
        identifier = kwargs.pop(self.identifier_key)
        model = self.store.model_class(id=identifier, **kwargs)
        return self.store.update(identifier, model)

    def update_batch(self, **kwargs):
        """
        Simplistic batch update operation implemented in terms of `replace()`.

        Assumes that:

         - Request and response schemas contains lists of items.
         - Request items define a primary key identifier
         - The entire batch succeeds or fails together.

        """
        items = kwargs.pop("items")

        def transform(item):
            """
            Transform the dictionary expected for replace (which uses the URI path's id)
            into the resource expected from individual resources (which uses plain id).

            """
            item[self.identifier_key] = item.pop("id")
            return item

        return dict(
            items=[
                self.replace(**transform(item))
                for item in items
            ],
        )
